# app.py
import os
import numpy as np
import streamlit as st
from deepface import DeepFace
from PIL import Image
import faiss
import pandas as pd

# Конфигурации
EMBEDDING_SIZE = 128  # Для FaceNet
DATABASE_PATH = "lab-4/task_2_faces"
INDEX_PATH = "lab-4/task_2_faiss_index.index"
NAMES_PATH = "lab-4/task_2_names.npy"
STAFF_CSV = "lab-4/task_2_staff_photo.csv"  # Добавим путь к CSV файлу

def build_faiss_database():
    st.write("Построение базы данных...")
    if not os.path.exists(DATABASE_PATH):
        st.error("Папка с лицами не найдена, создаю папку")
        st.write("Загрузите изображения лиц в папку и нажмите кнопку 'Построить/Обновить базу' ещё раз")
        os.makedirs(DATABASE_PATH)
        return

    # Загрузка данных о сотрудниках
    try:
        staff_df = pd.read_csv(STAFF_CSV)
        filename_to_name = dict(zip(staff_df['filename'], staff_df['name']))
    except Exception as e:
        st.error(f"Ошибка чтения CSV файла: {str(e)}")
        return

    embeddings = []
    names = []
    
    progress_bar = st.progress(0)
    files = sorted(os.listdir(DATABASE_PATH))
    mx = len(files)
    
    for i, img_file in enumerate(files):
        img_path = os.path.join(DATABASE_PATH, img_file)
        try:
            # Извлечение лица и эмбеддинга
            img = Image.open(img_path)
            face = DeepFace.extract_faces(img_path=img_path, detector_backend='opencv')[0]
            embedding = DeepFace.represent(
                img_path=img_path,
                model_name='Facenet',
                detector_backend='opencv'
            )[0]['embedding']
            
            embeddings.append(embedding)
            # Используем имя из CSV вместо имени файла
            person_name = filename_to_name.get(img_file, img_file)
            names.append(person_name)
        except Exception as e:
            st.error(f"Ошибка обработки {img_path}: {str(e)}")
        progress_bar.progress(i / mx)
    
    if len(embeddings) == 0:
        return

    # Создание FAISS индекса
    embeddings = np.array(embeddings).astype('float32')
    index = faiss.IndexFlatL2(EMBEDDING_SIZE)
    index.add(embeddings)
    
    # Сохранение индекса и имен
    faiss.write_index(index, INDEX_PATH)
    np.save(NAMES_PATH, np.array(names))
    st.success(f"База данных построена: {len(names)} лиц")

def search_face(query_img, threshold=0.5):
    if not os.path.exists(INDEX_PATH):
        st.error("Сначала постройте базу данных")
        return None, None

    try:
        # Конвертация PIL Image в numpy array
        query_img_array = np.array(query_img)
        
        # Извлечение эмбеддинга для запроса
        face = DeepFace.extract_faces(img_path=query_img_array, detector_backend='opencv')[0]
        query_embedding = DeepFace.represent(
            img_path=query_img_array,
            model_name='Facenet',
            detector_backend='opencv'
        )[0]['embedding']
        
        # Поиск в FAISS
        index = faiss.read_index(INDEX_PATH)
        names = np.load(NAMES_PATH)
        
        # Поиск ближайшего соседа
        query_embedding = np.array([query_embedding]).astype('float32')
        distances, indices = index.search(query_embedding, 1)
        
        if distances[0][0] <= threshold:
            return names[indices[0][0]], distances[0][0]
        else:
            return "Unknown", distances[0][0]
            
    except Exception as e:
        st.error(f"Ошибка распознавания: {str(e)}")
        return None, None

# Streamlit интерфейс
st.title("🔍 Face Recognition System")

# Секция для построения базы данных
st.sidebar.header("База данных лиц")
if st.sidebar.button("Построить/Обновить базу"):
    build_faiss_database()

# Секция для поиска
st.header("Поиск лица")
uploaded_file = st.file_uploader("Загрузите изображение лица", type=['jpg', 'png', 'jpeg'])

if uploaded_file is not None:
    image = Image.open(uploaded_file)
    st.image(image, caption="Загруженное изображение", use_container_width=True)
    
    if st.button("Найти совпадения"):
        name, distance = search_face(image, threshold=1000)
        if name:
            st.subheader(f"Результат: {name}")
            st.write(f"Расстояние: {distance:.4f}")
            if name == "Unknown":
                st.error("Совпадений не найдено в базе данных")
            else:
                st.success("Совпадение найдено!")
                # Загружаем данные о сотрудниках
                staff_df = pd.read_csv(STAFF_CSV)
                # Ищем запись сотрудника по имени
                staff_record = staff_df[staff_df['name'] == name]
                if not staff_record.empty:
                    # Получаем имя файла фотографии
                    photo_filename = staff_record.iloc[0]['filename']
                    photo_path = os.path.join(DATABASE_PATH, photo_filename)
                    if os.path.exists(photo_path):
                        matched_image = Image.open(photo_path)
                        st.image(matched_image, caption=f"Фотография сотрудника: {name}", 
                                use_container_width=True)