import cv2
import pandas as pd
import numpy as np
from deepface import DeepFace
from tqdm import tqdm
import os
import pickle
import hashlib

# Конфигурация
VIDEO_DIR = "lab-4/task_3_1_references/videos"
TIMECODE_CSV_DIR = "lab-4/task_3_1_references/labels"
REFERENCE_DIR = "lab-4/task_3_1_references/photos"
CACHE_FILE = "lab-4/task_3_1_cache/embeddings_cache.pkl"
DIR_HASH_FILE = "lab-4/task_3_1_cache/dir_hash.txt"
THRESHOLD = 0.1  # Порог косинусного сходства
FRAME_SKIP = 50  # Каждый 50-й кадр
FRAME_LIMIT = 20 * FRAME_SKIP  # Ограничение на количество кадров

video_files = [f for f in os.listdir(VIDEO_DIR) if f.endswith((".mp4", ".avi", ".mov"))]
print("Доступные видео:")
for idx, video in enumerate(video_files):
    print(f"{idx + 1}: {video}")

video_choice = int(input("Выберите номер видео для обработки: ")) - 1
VIDEO_PATH = os.path.join(VIDEO_DIR, video_files[video_choice])

# Загрузка соответствующего файла с таймкодами
TIMECODE_CSV = os.path.join(
    TIMECODE_CSV_DIR,
    video_files[video_choice].replace(video_files[video_choice].split(".")[-1], "csv"),
)


def get_dir_hash(directory):
    """Вычисляет хэш директории на основе имен файлов и их размеров"""
    hash_md5 = hashlib.md5()
    for root, _, files in os.walk(directory):
        for file in sorted(files):  # сортировка для стабильности хэша
            filepath = os.path.join(root, file)
            file_size = os.path.getsize(filepath)
            hash_md5.update(f"{filepath}{file_size}".encode())
    return hash_md5.hexdigest()


def normalize_embedding(embedding):
    return embedding / np.linalg.norm(embedding)


def load_reference_embeddings():
    # Проверяем существующий кэш
    current_hash = get_dir_hash(REFERENCE_DIR)
    if os.path.exists(CACHE_FILE) and os.path.exists(DIR_HASH_FILE):
        with open(DIR_HASH_FILE, "r") as f:
            cached_hash = f.read().strip()
        if cached_hash == current_hash:
            print("Загрузка эмбеддингов из кэша...")
            with open(CACHE_FILE, "rb") as f:
                return pickle.load(f)

    print("Вычисление новых эмбеддингов...")
    embeddings = {}
    for person in tqdm(os.listdir(REFERENCE_DIR)):
        person_dir = os.path.join(REFERENCE_DIR, person)
        if os.path.isdir(person_dir):
            for img_name in os.listdir(person_dir):
                img_path = os.path.join(person_dir, img_name)
                try:
                    image_embeddings = [
                            normalize_embedding(emb["embedding"])
                            for emb in DeepFace.represent(
                                img_path=img_path,
                                model_name="Facenet",
                                detector_backend="retinaface",
                                enforce_detection=False,
                            )
                        ]
                    for embedding in image_embeddings:
                        if person not in embeddings:
                            embeddings[person] = []
                        embeddings[person].append(embedding)
                except Exception as e:
                    print(f"Ошибка обработки {img_path}: {str(e)}")
                    continue

    # Сохраняем новый кэш
    with open(CACHE_FILE, "wb") as f:
        pickle.dump(embeddings, f)
    with open(DIR_HASH_FILE, "w") as f:
        f.write(current_hash)

    return embeddings


print("\nЗагрузка эталонных эмбеддингов...")
reference_embeddings = load_reference_embeddings()
print(f"Загружено {len(reference_embeddings)} эталонных личностей")

# Загрузка разметки
print("\nЗагрузка разметки из CSV...")
timecodes = pd.read_csv(TIMECODE_CSV)
timecodes["from"] = pd.to_timedelta(timecodes["from"]).dt.total_seconds()
timecodes["to"] = pd.to_timedelta(timecodes["to"]).dt.total_seconds()
# Преобразование строки persons в список
timecodes["persons"] = timecodes["persons"].apply(
    lambda x: [person.strip() for person in x.split(",")] if isinstance(x, str) else x
)
print(f"Загружено {len(timecodes)} временных меток")

# Инициализация видео
print("\nИнициализация видео...")
cap = cv2.VideoCapture(VIDEO_PATH)
fps = cap.get(cv2.CAP_PROP_FPS)
total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
total_frames = min(total_frames, FRAME_LIMIT) if FRAME_LIMIT else total_frames
print(f"FPS: {fps:.2f}")
print(f"Всего кадров: {total_frames}")
print(f"Длительность видео: {total_frames/fps:.2f} секунд")

# Метрики
true_positives = 0
false_positives = 0

print("\nНачинаем обработку кадров...")
# Обработка кадров
for frame_num in tqdm(range(0, total_frames, FRAME_SKIP)):
    cap.set(cv2.CAP_PROP_POS_FRAMES, frame_num)
    ret, frame = cap.read()
    if not ret:
        break

    # Определение текущего времени
    current_time = frame_num / fps

    # Получение ожидаемых лиц
    current_frames = timecodes[
        (timecodes["from"] <= current_time) & (timecodes["to"] >= current_time)
    ]
    expected = []
    for persons_list in current_frames["persons"]:
        if isinstance(persons_list, list):
            expected.extend(persons_list)
    expected = list(set(expected))  # удаление дубликатов

    try:
        # Обнаружение лиц
        detected_faces = DeepFace.extract_faces(
            frame, detector_backend="retinaface", enforce_detection=False
        )
    except Exception as exc:
        print(f"Поиск на кадре {frame_num} не удался: {exc=}")
        continue

    # print(f"{expected=}")

    for face in detected_faces:
        try:
            # Получение эмбеддинга
            query_embedding = normalize_embedding(
                DeepFace.represent(
                    img_path=face["face"],
                    model_name="Facenet",
                    detector_backend="retinaface",
                    enforce_detection=False,
                )[0]["embedding"]
            )

            # Поиск совпадений
            best_match = ("Unknown", 1)
            for person, embeddings in reference_embeddings.items():
                avg_distance = np.mean(
                    [np.linalg.norm(query_embedding - emb) for emb in embeddings]
                )
                if avg_distance < best_match[1]:
                    best_match = (person, avg_distance)
            # Проверка совпадения
            # print(f"{best_match=}")
            if best_match[1] > THRESHOLD:
                if best_match[0] in expected:
                    true_positives += 1
                else:
                    false_positives += 1
            else:
                false_positives += 1

        except Exception as exc:
            print(f"Сопоставление на кадре {frame_num} не удалось: {exc=}")
            false_positives += 1

# Расчет точности
precision = (
    true_positives / (true_positives + false_positives)
    if (true_positives + false_positives) > 0
    else 0
)
print("\nРезультаты анализа:")
print(f"Правильные обнаружения (True Positives): {true_positives}")
print(f"Ложные обнаружения (False Positives): {false_positives}")
print(f"Precision: {precision:.4f}")
cap.release()
print("\nОбработка завершена!")
