"""## Часть 2. Оценить точность работы методов из DeepFace на аугментированных данных (2 балла).

Необходимо собрать собственный набор данных из **различных** изображений Вашего лица с разных ракурсов, желательно настоящие фотографии из личного архива (20 штук)\
Возьмите эталонное изображение (как в паспорте) и при помощи библиотеки [DeepFace](https://github.com/serengil/deepface) проверьте его на соответствие всему датасету. Посчитайте метрику Precision. \
\
Примените каждую из перечисленных ниже аугментаций (**по-отдельности**) ко всему датасету и измерьте метрику Precision для измененнного датасета:
*   Поворот изображения на 45° и 90°.
*   Добавление шума (Gaussian Noise).
*   Изменение яркости (увеличение и уменьшение на 50%).
*   Размытие с различными параметрами.
\
Реузультаты соберите в таблицу вида:

Метод | Исходный датасет | Поворот на 45° | Поворот на 90° | Изображение с шумом |
--- | ----|--- | --- | --- |
VGG-Face | 0 | 0 | 0 | 0 |
Facenet | 0 | 0 | 0 | 0 |
Facenet512 | 0 | 0 | 0 | 0 |
OpenFace | 0 | 0 | 0 | 0 |
DeepFace | 0 | 0 | 0 | 0 |
DeepID | 0 | 0 | 0 | 0 |
ArcFace | 0 | 0 | 0 | 0 |
Dlib | 0 | 0 | 0 | 0 |
SFace | 0 | 0 | 0 | 0 |
GhostFaceNet | 0 | 0 | 0 | 0 |
"""

import os
import cv2
import numpy as np
from deepface import DeepFace
from PIL import Image, ImageEnhance, ImageFilter
import pandas as pd
import matplotlib.pyplot as plt
from tqdm import tqdm

# Путь к эталонному изображению
reference_image_path = "lab-4/task_3_2_reference.jpg"

# Путь к папке с изображениями датасета
dataset_folder = "lab-4/task_3_2_references"
image_paths = [os.path.join(dataset_folder, img) for img in os.listdir(dataset_folder)]

"""Поворот на 45 и 90 градусов:"""


def rotate_image(image, angle):
    (h, w) = image.shape[:2]
    center = (w // 2, h // 2)
    M = cv2.getRotationMatrix2D(center, angle, 1.0)
    rotated = cv2.warpAffine(image, M, (w, h))
    return rotated


"""Добавление шума:"""


def add_noise(image):
    row, col, ch = image.shape
    mean = 0
    var = 0.1
    sigma = var**0.5
    gauss = np.random.normal(mean, sigma, (row, col, ch))
    gauss = gauss.reshape(row, col, ch)
    noisy = image + gauss * 50
    return noisy.astype(np.uint8)


"""Изменение яркости:"""


def adjust_brightness(image, factor):
    img = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
    enhancer = ImageEnhance.Brightness(img)
    brightened = enhancer.enhance(factor)
    return cv2.cvtColor(np.array(brightened), cv2.COLOR_RGB2BGR)


"""Размытие:"""


def apply_blur(image, kernel_size=(5, 5)):
    return cv2.GaussianBlur(image, kernel_size, 0)


def apply_average_blur(image, kernel_size=(5, 5)):
    return cv2.blur(image, kernel_size)


def apply_bilateral_blur(image, d=9, sigma_color=75, sigma_space=75):
    return cv2.bilateralFilter(image, d, sigma_color, sigma_space)


"""Распознавание лиц:"""


def calculate_recall(reference_embedding, dataset_embeddings, threshold=0.5):
    print(f"Вычисление метрики recall с порогом {threshold}")
    true_positives = 0
    false_negatives = 0

    for i, embedding_list in enumerate(dataset_embeddings):
        if isinstance(embedding_list, list) and len(embedding_list) > 0:
            embedding_dict = embedding_list[0]
            if isinstance(embedding_dict, dict) and "embedding" in embedding_dict:
                embedding = np.array(embedding_dict["embedding"])
                distance = np.linalg.norm(reference_embedding - embedding)
                if distance < threshold:
                    true_positives += 1
                else:
                    false_negatives += 1
            else:
                print(
                    f"Предупреждение: Неожиданный формат embedding_dict в элементе {i+1}"
                )
                false_negatives += 1
        else:
            print(f"Предупреждение: Лицо не обнаружено в элементе {i+1}")
            false_negatives += 1

    recall = true_positives / 20
    print(
        f"Вычисление recall завершено, {recall=:.3f}. Истинно положительных: {true_positives}, Ложно отрицательных: {false_negatives}"
    )
    return recall


def test_transformations(image_paths):
    results = {
        model: {}
        for model in [
            "Facenet",
            "VGG-Face",
            "Facenet512",
            "OpenFace",
            # "DeepFace",
            "DeepID",
            "ArcFace",
            "Dlib",
            "SFace",
            "GhostFaceNet",
        ]
    }

    for model_name in results.keys():
        print(f"\nТестирование модели: {model_name}")

        # First extract face from reference image
        print("Извлечение лица из эталонного изображения...")
        try:
            faces = DeepFace.extract_faces(
                img_path=reference_image_path,
                detector_backend="opencv",
                enforce_detection=False,
            )

            if faces:
                # Get reference embedding using extracted face
                print("Получение эталонного эмбеддинга...")
                reference_embedding_list = DeepFace.represent(
                    img_path=faces[0]["face"],
                    model_name=model_name,
                    detector_backend="opencv",
                    enforce_detection=False,
                )
                reference_embedding = np.array(reference_embedding_list[0]["embedding"])
            else:
                print("Лицо не обнаружено в эталонном изображении")
                continue
        except Exception as e:
            print(f"Ошибка при обработке эталонного изображения: {str(e)}")
            continue

        print("Обработка исходного датасета...")
        original_embeddings = process_dataset(image_paths, model_name=model_name)
        results[model_name]["Исходный датасет"] = calculate_recall(
            reference_embedding, original_embeddings
        )

        print("Обработка поворота на 45°...")
        rotated_45_embeddings = process_dataset(
            image_paths,
            model_name=model_name,
            transform=lambda img: rotate_image(img, 45),
            transform_name="Поворот на 45°",
        )
        results[model_name]["Поворот на 45°"] = calculate_recall(
            reference_embedding, rotated_45_embeddings
        )

        print("Обработка поворота на 90°...")
        rotated_90_embeddings = process_dataset(
            image_paths,
            model_name=model_name,
            transform=lambda img: rotate_image(img, 90),
            transform_name="Поворот на 90°",
        )
        results[model_name]["Поворот на 90°"] = calculate_recall(
            reference_embedding, rotated_90_embeddings
        )

        print("Обработка изображений с шумом...")
        noisy_embeddings = process_dataset(
            image_paths,
            model_name=model_name,
            transform=add_noise,
            transform_name="Изображение с шумом",
        )
        results[model_name]["Изображение с шумом"] = calculate_recall(
            reference_embedding, noisy_embeddings
        )

        print("Обработка изображений с уменьшенной яркостью...")
        low_brightness_embeddings = process_dataset(
            image_paths,
            model_name=model_name,
            transform=lambda img: adjust_brightness(img, 0.5),
            transform_name="Изображение с уменьшенной яркостью",
        )
        results[model_name]["Изображение с уменьшенной яркостью"] = calculate_recall(
            reference_embedding, low_brightness_embeddings
        )

        print("Обработка изображений с увеличенной яркостью...")
        high_brightness_embeddings = process_dataset(
            image_paths,
            model_name=model_name,
            transform=lambda img: adjust_brightness(img, 1.5),
            transform_name="Изображение с увеличенной яркостью",
        )
        results[model_name]["Изображение с увеличенной яркостью"] = calculate_recall(
            reference_embedding, high_brightness_embeddings
        )

        print("Обработка изображений с размытием...")
        blurred_embeddings = process_dataset(
            image_paths,
            model_name=model_name,
            transform=apply_blur,
            transform_name="Размытие",
        )
        results[model_name]["Размытие"] = calculate_recall(
            reference_embedding, blurred_embeddings
        )

    return results


def process_dataset(image_paths, model_name, transform=None, transform_name="Без трансформации"):
    embeddings = []

    for img_path in tqdm(
        image_paths,
        desc=f"Обработка изображений ({model_name}, трансформация={transform_name})",
    ):
        try:
            img = cv2.imread(img_path)
            if transform:
                img = transform(img)

            # First extract face
            faces = DeepFace.extract_faces(
                img_path=img, detector_backend="opencv", enforce_detection=False
            )

            if faces:  # If face was found
                # Get embedding using the extracted face
                embedding_list = DeepFace.represent(
                    img_path=faces[0]["face"],  # Use the extracted face
                    model_name=model_name,
                    detector_backend="opencv",
                    enforce_detection=False,
                )
                embeddings.append(embedding_list)
            else:
                print(f"Лицо не обнаружено в {os.path.basename(img_path)}")
                embeddings.append([])
        except Exception as e:
            print(
                f"Ошибка обработки {os.path.basename(img_path)} моделью {model_name}: {str(e)}"
            )
            embeddings.append([])
    return embeddings


def create_results_table(results):
    # Convert results dictionary to DataFrame format
    data = {
        "Метод": list(results.keys()),
        "Исходный датасет": [results[model]["Исходный датасет"] for model in results],
        "Поворот на 45°": [results[model]["Поворот на 45°"] for model in results],
        "Поворот на 90°": [results[model]["Поворот на 90°"] for model in results],
        "Изображение с шумом": [
            results[model]["Изображение с шумом"] for model in results
        ],
        "Изображение с уменьшенной яркостью": [
            results[model]["Изображение с уменьшенной яркостью"]
            for model in results
        ],
        "Изображение с увеличенной яркостью": [
            results[model]["Изображение с увеличенной яркостью"] for model in results
        ],
        "Размытие": [results[model]["Размытие"] for model in results],
    }

    results_df = pd.DataFrame(data)
    return results_df


def main():
    print("Запуск оценки распознавания лиц...")
    print(f"Используется эталонное изображение: {reference_image_path}")
    print(f"Папка с датасетом: {dataset_folder}")

    results = test_transformations(image_paths)

    print("\nСоздание таблицы результатов...")
    results_df = create_results_table(results)
    print("\nТаблица результатов:")
    print(results_df)

    print("\nСохранение результатов в CSV...")
    results_df.to_csv("lab-4/task_3_2_results.csv", index=False)
    print("Результаты успешно сохранены")


if __name__ == "__main__":
    main()
